/*
 *  Project: LKH - lect9
 *  File: es1.c
 *  Date: 28/09/2005
 *  Author. M. Cesati, R. Gioiosa
 *
 *  12/03/06: Init errors
 *            debug macro
 */


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/timer.h>
#include <linux/vmalloc.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>
#include <asm/uaccess.h> 

#define MODULE_VERS "1.1"
#define MODULE_NAME "bldex"

#define DEBUG_ON

#ifdef DEBUG_ON
#define Dprintk(ARG...) printk(KERN_DEBUG MODULE_NAME ": " ARG)
#else
#define Dprintk(ARG...) 
#endif

#define DEVSIZE_MB 16
#define BLDEX_MINORS 16
#define FIRST_MINOR 0

struct bldex_desc_t {
        int major;
        int size_in_sectors;
        u8 *data;
        int users;
        spinlock_t lock;
        struct request_queue *queue;
        struct gendisk *gd;
};

static struct bldex_desc_t bldex_desc;

static int bldex_open(struct inode *inode, struct file *filp)
{
        struct bldex_desc_t *dev = inode->i_bdev->bd_disk->private_data;

        filp->private_data = dev;
        spin_lock(&dev->lock);
        if (! dev->users)
                check_disk_change(inode->i_bdev);
        dev->users++;
        spin_unlock(&dev->lock);
        return 0;
}

static int bldex_release(struct inode *inode, struct file *filp)
{
        struct bldex_desc_t *dev = inode->i_bdev->bd_disk->private_data;

        spin_lock(&dev->lock);
        dev->users--;
        if (!dev->users) {
                Dprintk("bldex: no other users left\n");
        }
        spin_unlock(&dev->lock);
        return 0;
}

static int bldex_ioctl(struct inode *inode, struct file *filp,
        unsigned int cmd, unsigned long arg)
{
        long size;
        struct hd_geometry geo;
        struct bldex_desc_t *dev = filp->private_data;

        switch(cmd) {
                case HDIO_GETGEO:
                /*
                * Get geometry: since we are a virtual device, we have to make
                * up something plausible. So we claim 16 sectors, four heads,
                * and calculate the corresponding number of cylinders. 
                * We set the start of data at sector four.
                */
                size = dev->size_in_sectors;
                geo.cylinders = (size & ~0x3f) >> 6;
                geo.heads = 4;
                geo.sectors = 16;
                geo.start = 4;
                if (copy_to_user((void __user *) arg, &geo, sizeof(geo)))
                        return -EFAULT;
                return 0;
        }
        return -ENOTTY; /* unknown command */
}

static struct block_device_operations bldex_fops = {
        .open = bldex_open,
        .release = bldex_release,
        .ioctl = bldex_ioctl,
};

static void bldex_transfer(struct bldex_desc_t *dev, unsigned long sector,
                                unsigned long nsect, char *buffer, int write)
{
        unsigned long offset = sector*512;
        unsigned long nbytes = nsect*512;

        if ((offset + nbytes) > 512*dev->size_in_sectors) {
                Dprintk("Beyond-end write (%ld %ld)\n",offset, nbytes);
                return;
        }
        if (write)
                memcpy(dev->data + offset, buffer, nbytes);
        else
                memcpy(buffer, dev->data + offset, nbytes);
}

static void bldex_request(request_queue_t *q)
{
        struct request *req;

        while ((req = elv_next_request(q)) != NULL) {
                struct bldex_desc_t *dev = req->rq_disk->private_data;
                if (! blk_fs_request(req)) {
                        Dprintk(KERN_NOTICE "bldex: skip non-fs request\n");
                        end_request(req, 0);
                        continue;
                }
                bldex_transfer(dev, req->sector, req->current_nr_sectors,
                        req->buffer, rq_data_dir(req));
                end_request(req, 1);
        }
}

static int __init bldex_init(void)
{
        int rc; 
        int size_in_bytes = DEVSIZE_MB * 1024 * 1024;
        
        printk(KERN_INFO "bldex: block device driver example\n");
        printk(KERN_INFO "%s v%s\n", MODULE_NAME, MODULE_VERS);
        
        memset(&bldex_desc, 0, sizeof(bldex_desc));
        bldex_desc.size_in_sectors = size_in_bytes/512;
        if (!(bldex_desc.data = vmalloc(size_in_bytes))) {
                printk(KERN_ERR "bldex: memory allocation error\n");
		goto err;
        }
        
        rc = register_blkdev(0, MODULE_NAME);
        if (rc < 0) {
                printk(KERN_ERR "blkdex: major number allocation failed\n");
		goto err_major;
        }
        bldex_desc.major = rc;
        
        spin_lock_init(&bldex_desc.lock);
        
        bldex_desc.queue = blk_init_queue(bldex_request, &bldex_desc.lock);
        if (bldex_desc.queue == NULL) {
                printk(KERN_ERR "bldex: queue allocation failed\n");
		goto err_queue;
        }
        
        if (!(bldex_desc.gd = alloc_disk(BLDEX_MINORS))) {
                printk(KERN_ERR "bldex: alloc_disk failed\n");
		goto err_disk;
        }
        
        blk_queue_hardsect_size(bldex_desc.queue, 512);
        bldex_desc.queue->queuedata = &bldex_desc;
        bldex_desc.gd->major = bldex_desc.major;
        bldex_desc.gd->first_minor = FIRST_MINOR;
        bldex_desc.gd->fops = &bldex_fops;
        bldex_desc.gd->queue = bldex_desc.queue;
        bldex_desc.gd->private_data = &bldex_desc;
        snprintf(bldex_desc.gd->disk_name, 32, "bldex%c", 'a'+FIRST_MINOR); 
        set_capacity(bldex_desc.gd, bldex_desc.size_in_sectors);
        add_disk(bldex_desc.gd);
        return 0;

 err_disk:
	blk_cleanup_queue(bldex_desc.queue);
 err_queue:
	unregister_blkdev(bldex_desc.major, MODULE_NAME);
 err_major:
	vfree(bldex_desc.data);
 err:
	return 1;
}

static void __exit bldex_exit(void)
{
	if (bldex_desc.gd) {
                del_gendisk(bldex_desc.gd);
                put_disk(bldex_desc.gd);
        }
	if (bldex_desc.queue)
               blk_cleanup_queue(bldex_desc.queue);
        if (bldex_desc.major) {
                int rc;
                rc = unregister_blkdev(bldex_desc.major, MODULE_NAME);
                if (rc < 0)
                        printk(KERN_ERR "blkdex: major number deallocation failed\n");
	}
	if (bldex_desc.data)
               vfree(bldex_desc.data);
        printk(KERN_INFO "bldex unloaded\n");
}

module_init(bldex_init);
module_exit(bldex_exit);

MODULE_AUTHOR("M.Cesati,R.Gioiosa");
MODULE_DESCRIPTION("Block device driver example");
MODULE_LICENSE("GPL");
MODULE_VERSION(MODULE_VERS);
